% Author: Ibrahim Ghafir
% Datasets: https://doi.org/10.17028/rd.lboro.4746844.v1
% Licence:CCBY 4.0 

% This file only observes TTL - no DS


%% This script is to assign the normality and abnormality beliefs for each 
%  sample of each metric in the test_data

% 1- the main data is divided into two partitions:
%  profile_data: used to build the profiles of normality and abnormality
%                beliefs (80% of the data). 
%  test_data: used to assign the beliefs, label the test_data based on the 
%  normality and abnormality beliefs, and test Dumster Shafer theory
%             (20% of the data).


% 2- To assign the normality belief for each sample of each metric, we have
% two steps:

% in the first one, we filter the significant outliers using k-means 
% clustering on all metrics of the profile_data

% in the second step, we assume that the cleaner profile_data is normally 
% distributed and use the Gaussian formula to find the mean (u) and standard 
% deviation (sigma) of the profile_data.

% for each sample of each metric of the test_data, we apply the Gaussian
% formula and calculate (pdf(i)) using the mean (u) and standard deviation 
% (sigma) of the profile_data

% to assign the normality belief, we scale the calculated value pdf(i) to 
% the pdf_max whic is pdf(u). pdf(u) is used as a reference.

% 3- To assign the abnormality belief for each sample of each metric, we
% are based on Local Reachability Density (lrd). 

% For each metric in the profile_data, we calculate the lrd for each
% sample, and find lrd_max.

% for each metric in the test_data, we calculate the lrd based on the
% profile_data

% To assign the abnormality belief, we scale the calculated lrd to lrd_max.
% lrd_max is used as a reference.

% 4- then we label the test_data based on the normality and abnormality 
% beliefs, and calculate TPR and FPR for each metric.

%

%  1- we have two datasets (mitm_attack01_data.txt) and
%  (mitm_attack02_data.txt), first we can load both of them and then
%  assigne x to one of them to identifiy its outlier cluster.
%  2- label the data, based on TTl = 255 and Injection rate = 1
%  3- change the TTL values (from 255 to 50) and change the sequence number
%  4- apply kmeans clustering based on all metrics of the frame.
%  5- apply kmeans clustering based on 3 metrics of the frame.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%  1- load the data and assign x to the test metrics.

load mitm_attack01_data.txt

load mitm_attack02_data.txt

x = mitm_attack02_data (:,:);

% 2- label the data, based on TTl = 255 and Injection rate = 1

for i=1:size(x,1)  % size(x,1) is the number of rows in the matrix x
    if (x(i,3)==255) && (x(i,2)==1)
        x(i,7)=1;
    else
         x(i,7)=0;
    end
end

% 3- change the TTL values (from 255 to 50) and change the sequence number

% change TTL value

for i=1:size(x,1)  % size(x,1) is the number of rows in the matrix x
    if (x(i,3)==255)
        x(i,3)=50;
    end
end

% change sequence number value

for i=size(x,1):-1:2  % size(x,1) is the number of rows in the matrix x
    x(i,5)=x(i,5)-x(i-1,5);
end

x(1,5)=0;

%% partitioning the data into:
%  profile_data: used to build the profiles of normality and abnormality
%                beliefs (80% of the data). 
%  test_data: used to assign the beliefs and test Dumster Shafer theory
%             (20% of the data).


ttl_small_window_results = [];

for x_1=0:10

sub_x = x((x_1*1000)+1:(x_1+1)*1000,:);

profile_data=sub_x(1:end*0.8,:);
test_data=sub_x(end*0.8+1:end,:);

%profile_data=x(1:end*0.8,:);
%test_data=x(end*0.8+1:end,:);

%test_data=x; % This is to evaluate the algorithms on all data (test data 100%)


%% Normality belief assignment (gaussian distribution)

% kmeans clustering

%  apply kmeans clustering based on all metrics of the frame, to filter the
%  significant outliers, and to get more accurate normality profile

kmeans_data = profile_data(:,1:6);

% this is to plot the data (only 2 dimensions)
%figure(1);
%plot(kmeans_data(:,1),kmeans_data(:,2),'.');

opts = statset('Display','final');
[idx,C] = kmeans(kmeans_data,2,'Distance','cityblock',...
    'Replicates',5,'Options',opts);

%opts = statset('Display','final');
%[idx,C] = kmeans(kmeans_data,2);

% this is to plot the clusered data in different coulours, besides the
% centroids

%figure(2);
%plot(kmeans_data(idx==1,1),kmeans_data(idx==1,2),'r.','MarkerSize',12)
%hold on
%plot(kmeans_data(idx==2,1),kmeans_data(idx==2,2),'b.','MarkerSize',12)
%plot(C(:,1),C(:,2),'kx',...
%     'MarkerSize',15,'LineWidth',3)
%legend('Cluster 1','Cluster 2','Centroids',...
%       'Location','NW')
%title 'Cluster Assignments and Centroids'
%hold off

% to see the number of elements of each cluster
%figure(1);
%histogram (idx);

% label the data based on kmeans clustring, considering the larger cluster
% is the normal one, and the smaller cluster is the outliers.

% kmeans define each cluster with a number, here we have 2 clusters, so
% each alement in the data is clustered into '1' or '2'.

% to find the most frequent value in idx (the larger cluster) 

normal_cluster = mode(idx);

for i=1:size(profile_data,1)  % size(x,1) is the number of rows in the matrix x
    if (idx(i,1)==normal_cluster)
        profile_data(i,8)=0;
    else
         profile_data(i,8)=1;
    end
end 

cleaner_data=profile_data;

outliers_rows = find(cleaner_data(:,8)==1);
cleaner_data(outliers_rows,:) = [];

%%% Different distributions are to be utilised for normality belief assignment:

%Gaussian distribution: for RSSI, seq, TTL
%Categorical distribution (Bernoulli): for Rate, NAV
%Exponential distribution: for IAT
%%%%%%%%

% Gaussian distribution and normality belief assignment
% we build the normality profile based on the cleaner_data (calculate u and
% sigma) and then used the previous values to apply gaussian formula on the
% test_dada to find the normality belief.

% for now, lets consider al the metrics are gaussian distributed, however,
% we might need to consider the bernoli distribution for 2 and 4 metrics.

% for Gaussian distribution

for metric = [3];

gaussian_data = [];
normality_belief = [];

gaussian_data = cleaner_data(:,metric);

normality_belief = (normpdf(test_data(:,metric),mean(gaussian_data),std(gaussian_data)) ...
    / normpdf(mean(gaussian_data),mean(gaussian_data),std(gaussian_data)))*100;

test_data(:,metric+10) = normality_belief;

end

%{
% for Categorical distribution

for metric = [2 4];

categorical_data = [];
normality_belief = [];
categorical_data = cleaner_data(:,metric);
categorical_matrix = [];

categorical_unique = unique(categorical_data);

for i = 1:length(categorical_unique)
occurrences(i) = sum(categorical_data==categorical_unique(i));
frequency(i) = (occurrences(i) / length(categorical_data))*100;
% the folowing matrix gives the normlity belief ofeach unique value
categorical_matrix = [categorical_matrix; categorical_unique(i), frequency(i)];
end

for i=1:size(test_data,1)
    y=ismember(test_data(i,metric),categorical_matrix(:,1));
        if y==0
            normality_belief(i) = 0;
        else
            for j=1:size(categorical_matrix,1)
                if test_data(i,metric) == categorical_matrix(j,1)
                    normality_belief(i) = categorical_matrix(j,2);
                end
            end
        end
end

test_data(:,metric+10) = normality_belief;

end

% for exponential distribution

% the follwoing 3 lines are for debugging purpose
exp_metric_mu = [];
exp_metric_min = [];
exp_metric_P_max = [];

for metric = 6;

exponential_data = [];
normality_belief = [];

exponential_data = cleaner_data(:,metric);
exponential_mean = mean(exponential_data);

exp_metric_mu(metric) = exponential_mean; % for debugging purpose
metric_min = min(exponential_data);
P_max = exppdf(metric_min,exponential_mean);

 for i=1:size(test_data,1)
     if test_data(i,metric) < metric_min
        metric_min = test_data(i,metric);
        P_max = exppdf(metric_min,exponential_mean);
     end
    
   normality_belief(i) = (exppdf(test_data(i,metric),exponential_mean) / P_max)*100;

 end
 
 exp_metric_min(metric) = metric_min;
 exp_metric_P_max(metric) = P_max;
 test_data(:,metric+10) = normality_belief;

end

%}
%% Abnormality belief assignment (Local Reachability Density)

% For each metric in the profile_data, we calculate the lrd for each
% sample, and find lrd_max. This includes using the k-nearest algorithm

% for each metric in the test_data, we calculate the lrd based on the
% profile_data. This includes using the k-nearest algorithm

% To assign the abnormality belief, we scale the calculated lrd to lrd_max.
% lrd_max is used as a reference.

% 1- find the k-nearest neighbours of each sample (frame) in the 
% profile_data

% keep only the unique values in the knn data, to avoid the zero distances
% resulting from the dublicator values.


% think of calculation the local reachability densit of all samples,
% then chose the max as the 100% normal
% then calulate the local reaxhability desity of your test sample
% then scale to the max lrd
% abnornality beief = [1-(lrd/lrd_max)]*100


% assign k, based on the number of unique values

for metric=[3];

knn_data=[];

knn_data=unique(profile_data(:,metric));

if size(knn_data,1)<21
    k=round(0.7*size(knn_data,1)); 
else
    k=20;
end

for i=1:size(knn_data,1) % size(x,1) is the number of rows in the matrix x
    knn_data_i=[];
    neigbours_p=[];
    distances_p=[];
   % k_dist_p=[];
    
    knn_data_i=knn_data;
    p=knn_data_i(i);
    knn_data_i(i)=[]; % to avoid the zero distance between identical samples
   
    [neigbours_p,distances_p]=knnsearch(knn_data_i, p,'k',k);
    %k_dist_p=distances_p(k); % not required for calculating lrd_k_p
    
    k_dist_o=[];
    dist_p_o=[];
    reach_dist_p_o=[];
    
    for j=1:size(neigbours_p,2)
        knn_data_i_j=[];
        neigbours_o=[];
        distances_o=[];
        
        knn_data_i_j=knn_data_i;
        o=knn_data_i_j(neigbours_p(j));
        knn_data_i_j(neigbours_p(j))=[]; % delete the value of o, to avoid the duplicator samples
        knn_data_i_j=[knn_data_i_j;p]; % return the value of p previously deleted
           % and add it t the end of the matrix so you don't affect the
           % indexes of neigbours_p when they are used in the next for loop.
    
        [neigbours_o,distances_o]=knnsearch(knn_data_i_j, o,'k',k);
        k_dist_o(j)=distances_o(k);
        dist_p_o(j)=distances_p(j);
        %neighbours(j,1:k)=neigbours_o; % not required for calculating lrd_k_p
        reach_dist_p_o(j)=max(k_dist_o(j),dist_p_o(j));

    end
    
    lrd_p(i)=1/(sum(reach_dist_p_o)/k); % lrd for the profile_data sample

end

lrd_max(metric)= max(lrd_p);

lrd_min(metric)= min(lrd_p);

end

% for the test_data

for metric=[3];

knn_data=[];

knn_data=unique(profile_data(:,metric));

if size(knn_data,1)<21
    k=round(0.7*size(knn_data,1)); 
else
    k=20;
end

for i=1:size(test_data,1) % size(x,1) is the number of rows in the matrix x
    knn_data_i=[];
    neigbours_p=[];
    distances_p=[];
   % k_dist_p=[];
    
    knn_data_i=knn_data;
    p=test_data(i, metric);
    
    if ismember(p, knn_data_i)==1
        p_exist=1;
        knn_data_i(find(knn_data_i==p))=[]; % to delete if there is a duplicate value
    else
        p_exist=0;
    end
   
    [neigbours_p,distances_p]=knnsearch(knn_data_i, p,'k',k);
    %k_dist_p=distances_p(k); % not required for calculating lrd_k_p
    
    k_dist_o=[];
    dist_p_o=[];
    reach_dist_p_o=[];
    
    for j=1:size(neigbours_p,2)
        knn_data_i_j=[];
        neigbours_o=[];
        distances_o=[];
        
        knn_data_i_j=knn_data_i;
        o=knn_data_i_j(neigbours_p(j));
        knn_data_i_j(neigbours_p(j))=[]; % delete the value of o, to avoid the duplicator samples
        
        if p_exist==1
            knn_data_i_j=[knn_data_i_j;p]; % return the value of p previously deleted
        end
    
        [neigbours_o,distances_o]=knnsearch(knn_data_i_j, o,'k',k);
        k_dist_o(j)=distances_o(k);
        dist_p_o(j)=distances_p(j);
        %neighbours(j,1:k)=neigbours_o; % not required for calculating lrd_k_p
        reach_dist_p_o(j)=max(k_dist_o(j),dist_p_o(j));

    end
    
    lrd_p_t(i)=1/(sum(reach_dist_p_o)/k);  % lrd for the test_data sample
    
    if lrd_p_t(i) < lrd_min(metric)  % update lrd_min
        lrd_min(metric)=lrd_p_t(i);
    end
    
    if lrd_p_t(i) > lrd_max(metric)  % update lrd_max
        lrd_max(metric)=lrd_p_t(i);
    end
    
  %  abnormality_belief_1(i)=(lrd_p_t(i)*100)/lrd_min(metric);
    
  %  abnormality_belief_2(i)=1/((lrd_p_t(i)*100)/lrd_min(metric));
    
    abnormality_belief_3(i)=100-((lrd_p_t(i)*100)/lrd_max(metric));
    
    
    
  %  test_data(i,metric+20) = abnormality_belief_1(i);
    
   % test_data(i,metric+30) = abnormality_belief_2(i);
    
    test_data(i,metric+40) = abnormality_belief_3(i);

end

end

%% label the test_data based on Normality and abnormality beliefs

% lrd_max test

label=7;

for metric=[3];

for i=1:size(test_data,1)
   if test_data(i,metric+10) > test_data(i,metric+40)
       test_data(i, metric+50) = 0;
   else 
       test_data(i, metric+50) = 1;
   end
end

% calculate P, N, TP, TN, FP, FN, TPR, FPR

P=0; N=0; TP=0; TN=0; FP=0; FN=0;

for i=1:size(test_data,1)
    if test_data(i,label)==1
        P=P+1;
    else N=N+1;
    end
end

for i=1:size(test_data,1)
    if test_data(i,metric+50)==1
        if test_data(i,label)==1
            TP=TP+1;
        else FP=FP+1;
    end
    end
end

for i=1:size(test_data,1)
    if test_data(i,metric+50)==0
        if test_data(i,label)==0
            TN=TN+1;
        else FN=FN+1;
        end
    end
end

TPR = (TP/P)*100;
FPR = (FP/N)*100;
OSR = ((TP+TN)/(TP+FP+TN+FN))*100;
Precision = (TP/(TP+FP))*100;
F_score = 2*(((Precision/100)*(TPR/100))/((Precision/100)+(TPR/100)));

%disp (P); disp (N); disp (TP); disp (TN); disp (FP); disp (FN); disp (TPR); disp (FPR);

% write the results intp (iqr_results) matrix

beliefs_results(metric,1)=metric; beliefs_results(metric,2)=P; beliefs_results(metric,3)=N;
beliefs_results(metric,4)=TP; beliefs_results(metric,5)=TN; beliefs_results(metric,6)=FP;
beliefs_results(metric,7)=FN; beliefs_results(metric,8)=TPR; beliefs_results(metric,9)=FPR;
beliefs_results(metric,10)=OSR; beliefs_results(metric,11)=Precision; beliefs_results(metric,12)=F_score;

end

%%

% lrd_min test  % this test doesn't give good results, could be deleted

%{

label=7;

for metric=1:6

for i=1:size(test_data,1)
   if test_data(i,metric+10) > test_data(i,metric+30)
       test_data(i, metric+60) = 0;
   else 
       test_data(i, metric+60) = 1;
   end
end

% calculate P, N, TP, TN, FP, FN, TPR, FPR

P=0; N=0; TP=0; TN=0; FP=0; FN=0;

for i=1:size(test_data,1)
    if test_data(i,label)==1
        P=P+1;
    else N=N+1;
    end
end

for i=1:size(test_data,1)
    if test_data(i,metric+60)==1
        if test_data(i,label)==1
            TP=TP+1;
        else FP=FP+1;
    end
    end
end

for i=1:size(test_data,1)
    if test_data(i,metric+60)==0
        if test_data(i,label)==0
            TN=TN+1;
        else FN=FN+1;
        end
    end
end

TPR = (TP/P)*100;
FPR = (FP/N)*100;

%disp (P); disp (N); disp (TP); disp (TN); disp (FP); disp (FN); disp (TPR); disp (FPR);

% write the results intp (iqr_results) matrix

beliefs_results_2(metric,1)=metric; beliefs_results_2(metric,2)=P; beliefs_results_2(metric,3)=N;
beliefs_results_2(metric,4)=TP; beliefs_results_2(metric,5)=TN; beliefs_results_2(metric,6)=FP;
beliefs_results_2(metric,7)=FN; beliefs_results_2(metric,8)=TPR; beliefs_results_2(metric,9)=FPR;

end

%}

%% this is to continue the for loop for TTL small window (1000 samples each)

ttl_small_window_results = [ttl_small_window_results; beliefs_results(metric,:)];

end

ttl_small_window_results_mean = mean(ttl_small_window_results);



